from __future__ import absolute_import, division, print_function

import os
import random
import shutil
import sys
import time
import pdb

import numpy as np
import tensorflow as tf


from data_utils import BOS, EOS
from data_utils import load_dataset
from data_utils import create_batched_dataset
from data_utils import sentence_to_tensor
from models import model

from utils import DEFINE_boolean
from utils import DEFINE_integer
from utils import DEFINE_string
from utils import DEFINE_float
from utils import print_user_flags
from utils import loss_function, loss_function_with_label_smoothing

flags = tf.compat.v1.flags
FLAGS = flags.FLAGS

DEFINE_boolean("reset_output_dir", False, "Delete output_dir if exists.")
DEFINE_boolean("restore_checkpoint", True, "Auto retrieve checkpoint or not.")
DEFINE_boolean("is_toy", False, "Toy Dataset or not, useful for debugging")
DEFINE_boolean("is_attention", False, "Using Attention for Decoder or not")
DEFINE_boolean("translate_only", False, "Do translating for target to source only")
DEFINE_string("data_path", "en-fr", "Path to parallel corpus data (.txt)")
DEFINE_string("output_dir", "output", "Path to log folder")
DEFINE_integer("max_ckpt_to_keep", 3, "Max number of checkpoint to save")
DEFINE_integer("n_epochs", 10, "Number of training epochs")
DEFINE_integer("n_enc_layers", 2, "Number of stacked encoder RNN layers")
DEFINE_integer("n_dec_layers", 2, "Number of stacked decoder RNN layers")
DEFINE_integer("n_hidden", 256, "Dimensionality of RNN output")
DEFINE_integer("emb_dim", 256, "Dimensionality of word embedding, src==tgt")
DEFINE_integer("save_every", 500, "How many batches to save once")
DEFINE_integer("eval_every", 100, "How many batches to evaluate")
DEFINE_integer("batch_size", 64, "Batch size. SET to `2` for easy debugging.")
DEFINE_integer("n_loaded_sentences", 50000, "Number of sentences to load, "
                                            "Set to <= 0 for loading all data,"
                                            "SET LOWER FOR DEBUGGING")
DEFINE_float("init_lr", 1e-3, "Init learning rate. This is default for Adam.")
DEFINE_float("drop_keep_prob", 1.0, "Dropout rate")


def train():
  print("-" * 80)
  if not os.path.isdir(FLAGS.output_dir):
    print("Path {0} does not exist. Creating.".format(FLAGS.output_dir))
    os.makedirs(FLAGS.output_dir)
  elif FLAGS.reset_output_dir:
    print("Path {0} exists. Remove and remake.".format(FLAGS.output_dir))
    shutil.rmtree(FLAGS.output_dir)
    os.makedirs(FLAGS.output_dir)

  print_user_flags()

  # --------------------------------
  #             DATASET
  # --------------------------------
  data_dict = load_dataset(FLAGS.data_path,
                           is_toy=FLAGS.is_toy,
                           num_examples=FLAGS.n_loaded_sentences)
  batched_datasets = create_batched_dataset(data_dict,
                                            batch_size=FLAGS.batch_size)

  # --------------------------------
  #             MODEL
  # --------------------------------
  nmt_model = model(data_dict,
                    checkpoint_dir=FLAGS.output_dir,
                    emb_dim=FLAGS.emb_dim,
                    n_hidden=FLAGS.n_hidden,
                    n_enc_layers=FLAGS.n_enc_layers,
                    n_dec_layers=FLAGS.n_dec_layers,
                    max_ckpt_to_keep=FLAGS.max_ckpt_to_keep,
                    lr=FLAGS.init_lr)
  encoder, decoder = nmt_model['encoder'], nmt_model['decoder']

  # --------------------------------
  #             RESTORE
  # --------------------------------
  # try to restore saved model
  if FLAGS.restore_checkpoint:
    try:
      nmt_model['train_ckpt']. \
        restore(nmt_model['ckpt_manager'].latest_checkpoint).assert_consumed()
      print("\n\nTRYING TO RESTORE THE LATEST MODEL FROM: "
            "'{}'\n\n".format(FLAGS.output_dir), flush=True)
    except Exception as e:
      print("Error in restoring checkpoint: {}. Partially loaded "
            .format(e), flush=True)

  # --------------------------------
  #   TRANSLATE ONLY - NO TRAINING
  # --------------------------------
  # translate target to source
  if FLAGS.translate_only:
    print('translation started!!', flush=True)
    test_lines = open(os.path.join(FLAGS.data_path, 'target.txt')).\
        read().strip().split('\n')
    with open(os.path.join(FLAGS.data_path, 'translated_ss.txt'),'w') as f:
      for line in test_lines:
        result, _ = evaluate(line, encoder, decoder, data_dict,
                                  FLAGS.is_toy, FLAGS.is_attention)
        f.write(' '.join(result.split()[:-1]) + '\n')
    print('translation fininshed!!', flush=True)
    return

  # --------------------------------
  #             TRAIN
  # --------------------------------
  # (continue to) train now
  for epoch in range(1, FLAGS.n_epochs + 1):
    start = time.time()
    total_loss = 0
    total_ppl = 0


    for (batch_num, (src, tgt)) in enumerate(batched_datasets['train']):
      loss = 0
      with tf.GradientTape() as tape:
        loss = teacher_forcing(data_dict,
                               encoder,
                               decoder,
                               loss,
                               src,
                               tgt)# forward
      batch_loss = (loss / int(tgt.shape[1]))
      batch_ppl = np.exp(batch_loss)
      total_loss += batch_loss
      total_ppl += batch_ppl

      # backward
      variables = encoder.variables + decoder.variables
      gradients = tape.gradient(loss, variables)
      nmt_model['optim'].apply_gradients(zip(gradients, variables))

      if batch_num % 100 == 0:
        print('Epoch {} Batch {} Loss {:.4f} '
              'PPL {:.2f}'.format(epoch,
                                  batch_num,
                                  batch_loss.numpy(),
                                  batch_ppl),flush=True)
      # evaluate some sentences
      if (batch_num + 1) % FLAGS.eval_every == 0:
        sentences = draw_random_sentences(os.path.join(FLAGS.data_path,
                                                       'val.txt'),
                                          size=2)
        for sentence in sentences:
          translate(sentence, encoder, decoder, data_dict, FLAGS.is_toy,
                    FLAGS.is_attention, False)

      # saving (checkpoint) the model every 2 epochs
      if (batch_num + 1) % FLAGS.save_every == 0:
        save_path = nmt_model['ckpt_manager'].save()
        print("Model saved to '{}".format(save_path), flush=True)

    avg_ppl = total_ppl / (batch_num + 1)
    avg_loss = total_loss / (batch_num + 1)
    print('Epoch {} Loss {:.4f} PPL {:.2f}'.format(epoch,
                                                   avg_loss,
                                                   avg_ppl),flush=True)

    print('Time taken for this epoch {:.2f} sec\n'.format(time.time() - start),flush=True)
    print("-" * 80, flush=True)

    # final test after training
    # you can inject this into the loop to do it at every epoch's end
    if epoch == FLAGS.n_epochs:
      print("Testing the whole test set for epoch {}".format(epoch),flush=True)
      test_lines = open(os.path.join(FLAGS.data_path, 'target.txt')).\
        read().strip().split('\n')
      with open(os.path.join(FLAGS.data_path, 'translated_{}.txt'.format(epoch)),
                'w') as f:
        for line in test_lines:
          result, _ = evaluate(line, encoder, decoder, data_dict,
                                  FLAGS.is_toy, FLAGS.is_attention)
          f.write(' '.join(result.split()[:-1]) + '\n')


def teacher_forcing(data_dict, encoder, decoder, loss, src, tgt):
  """
  Training using Teacher-Forcing mode, accumulate loss function at every 
  decoding timestep using loss_function().  
  
  Remember the very first size of `src` and `tgt` is FLAGS.batch_size  
  
  Args:
    data_dict: for converting WORDS to INDEXES, needing for pre-padding BOS 
                at the decoding time step 0 
    encoder: 
    decoder: 
    loss: input here should be all zeros, then get accumulated at EVERY 
          timestep 
    src: batch of source sentences 
    tgt: batch of target sentences 

  Returns:
    loss value accumulated (don't average yet) 

  """
  # get encoder output
  keys, values = encoder(src)

  # TODO: initialize decoder hidden
  # make use of FLAGS.n_hidden, FLAGS.n_dec_layers, FLAGS.batch_size
  dec_hidden = None

  # insert BOS token to this batch for every sentence (dimension B x L x N)
  BOS_idx = data_dict['src_lang'].word2idx[BOS.strip()]
  dec_input = tf.expand_dims([BOS_idx] * FLAGS.batch_size, 1)

  # At every timestep, feed GROUND TRUTH not predicted token
  for t in range(1, tgt.shape[1]):
    # TODO: inference and calculate the loss for each token
    # remember to do "teacher" mode properly

  return loss


def draw_random_sentences(data_path, size=2):
  """ Draw random sentences from val.txt for evaluation"""
  lines = open(data_path).read().splitlines()

  sentences = []
  for _ in range(size):
    sentences.append(random.choice(lines).strip().split('\t')[-1])

  return sentences


def evaluate(sentence,
             encoder,
             decoder,
             data_dict,
             is_toy,
             is_attention=False):
  """
  Evaluaion, no "teacher" is allowed here. 
  
  Args:
    sentence: input to translate, from 'target.txt' 
    encoder: 
    decoder: 
    data_dict: 
    is_toy: 
    is_attention: 

  Returns:
    result: decoded sentence in language (not numbers), separated by spaces 
    sentence: untouched input sentence 

  """
  # convert sentence to tensor
  inputs_tensor = sentence_to_tensor(data_dict, is_toy, sentence)

  # get encoder output
  keys, values = encoder(inputs_tensor)

  # TODO: initialize decoder hidden
  dec_hidden = None

  # very first step is feeding BOS index to decoder to generate predictions
  BOS_idx = data_dict['tgt_lang'].word2idx[BOS.strip()]
  dec_input = tf.expand_dims([BOS_idx], 0)

  result = ''
  for _ in range(data_dict['max_length_tgt']):
    # TODO: furnish these
    predictions, dec_hidden = decoder(...)
    predicted_id = None
    result = None

    # end if predicts the EOS token
    if data_dict['tgt_lang'].idx2word[predicted_id] == EOS.strip():
      return result, sentence

    # the predicted ID is fed back into the model because we have no teacher
    dec_input = tf.expand_dims([predicted_id], 0)

  return result, sentence


def translate(sentence,
              encoder,
              decoder,
              data_dict,
              is_toy=False,
              is_attention=False,
              is_plot=False):
  result, sentence = evaluate(sentence,
                              encoder,
                              decoder,
                              data_dict,
                              is_toy,
                              is_attention)

  print('\tInput:\t\t{}'.format(sentence),flush=True)
  print('\tPredicted:\t\t{}'.format(result),flush=True)



if __name__ == "__main__":
  train()
